#------------------------------------------------------------------
# $Copyright (c) 2018 Veritas Technologies LLC.  All rights reserved. Veritas
# and the Veritas Logo are trademarks or registered trademarks of Veritas
# Technologies LLC or its affiliates in the U.S. and other countries. Other
# names may be trademarks of their respective owners.
#
# The Licensed Software and Documentation are deemed to be commercial computer
# software as defined in FAR 12.212 and subject to restricted rights as defined
# in FAR Section 52.227-19 "Commercial Computer Software - Restricted Rights" and
# DFARS 227.7202, et seq. "Commercial Computer Software and Commercial Computer
# Software Documentation," as applicable, and any successor regulations, whether
# delivered by Veritas as on premises or hosted services. Any use, modification,
# reproduction release, performance, display or disclosure of the Licensed Software
# and Documentation by the U.S. Government shall be solely in accordance with the
# terms of this Agreement. $
# 
# $Header: /project/isv-agent/VProAgents/PostgreSQL/Replication.pm,v 1.20 2015/06/18 11:30:42 mrampura Exp $ 
#
# $Author: mrampura $
# $Date: 2015/06/18 11:30:42 $
# $Revision: 1.20 $
#
# Provides:                            Replication
# Author:                              mrampura
# crAgent Build:                       1.88 2008/10/13 08:28:02
# crAgent Run Date:                    13 Oct, 2008
# Modification History:
# Known bugs/caveats:
#------------------------------------------------------------------

#------------
# Be strict..
#------------
use strict;
use warnings;

#----------------------
# Declare the package..
#----------------------
package PGReplication;

use ag_i18n_inc;

#-------------------------------------
# Include required framework modules..
#-------------------------------------
use Sys;            # Support generalized UNIX commands
use Proc;           # For calls to the process routines
use Comms;          # For calls to socket(3C) routines
use Version;        # Version display generator
use VCSagentFW;     # Primary Agent Interface to the ACC Library
use VCSDefines;     # This contains constants that VCS uses

#--------------------------------
# Include standard PERL modules..
#--------------------------------
use Fcntl;
use Socket;
use POSIX qw ( strftime );
use File::Copy;
use File::Path qw ( mkpath );
use File::stat;

#-----------------------------------------------------------------------
# Add our version number to the version hash and check if versions are
# requested..
#-----------------------------------------------------------------------
my $sVer = "\$Id: Replication.pm,v 1.20 2018/05/18 11:30:42 mrampura Exp $::ETVHeader";
Version::TellVersion ( $sVer );


my ( $CLUS_HOME, $sPrefix ) = ();

#------------------------
# Setup VCS/VCSone home..
#------------------------
if ( $ENV { CLUSTER_HOME } ) { # VCS 5.x/VCSOne..
  $CLUS_HOME ||= $ENV { CLUSTER_HOME } ||= "/opt/VRTSvcsone";
}
else { # VCS 4.x..
  $CLUS_HOME ||= $ENV { VCS_HOME } ||= "/opt/VRTSvcs";
}

( $sPrefix ) = $CLUS_HOME =~ /.*(vad|vcs(|one))$/;

#-----------------------------------------
# Define module-vendor specific commands..
#-----------------------------------------
my $sCmdHares  = $HARES  ||= "$CLUS_HOME/bin/hares";
my $sCmdHatype = $HATYPE ||= "$CLUS_HOME/bin/hatype";
my $sCmdHagrp  = $HAGRP  ||= "$CLUS_HOME/bin/hagrp";
my $sCmdHasys  = $HASYS  ||= "$CLUS_HOME/bin/hasys";
my $sCmdSSH = "/usr/bin/ssh";
my $sCmdCP = "/usr/bin/cp";
my $sCmdTouch = "/usr/bin/touch";


=pod
 -------------------------------------------------------------------------
 Function : int Rewind ( IN String BaseDir,
                         IN String DataDir,
                         IN Integer Port,
                         IN String HostName,
                         IN String PostgreSQLUser,
                         IN String SourceConnStr,
                         IN String RecoveryFile,
                         IN String EnvFile,
                         IN Integer TillWhen);

 Description : Uses PostgreSQL command pg_rewind to connect as standby 
 
 Parameters  :

 $sBaseDir                  The Installation Directory of PostgreSQL.
 $sDataDir                  Location of the database storage area.
 $iPort                     The port number on which PostgreSQL instance is configured.
 $sHostName                 Hostname/IPv4/IPv6 address of virtual host for PostgreSQL Instance
 $iPort                     The port number on which PostgreSQL instance is configured.
 $sPostgreSQLUser           Specifies the UserName required to run into pg_ctl.
 $sSourceConnStr            Connection string used by pg_rewind command
 $sRecoveryFile             recovery.conf file provided to start standby instance
 $sEnvFile                  Absolute path to environment file to source
 $iTillWhen                 Time in EPOCH left for Operation

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut




sub Rewind {
  #----------------
  # Get Arguments..
  #----------------
  my($sBaseDir, $sDataDir, $iPort, $sHostName, $sPostgreSQLUser, $sSourceConnStr, $sRecoveryFile, $sEnvFile, $iTillWhen) = @_;
  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];
  
  &logTrace ( 20392, "Subroutine <%s> called with
  BaseDir            [%s]
  DataDir            [%s]
  Port               [%s]
  HostName           [%s]
  PostgreSQLUser     [%s]
  SourceConnStr      [%s]
  RecoveryFile       [%s]
  EnvFile            [%s]
  TillWhen           [%s]",
  $sMyName, $sBaseDir, $sDataDir, $iPort, $sHostName, $sPostgreSQLUser, $sSourceConnStr, $sRecoveryFile, $sEnvFile, $iTillWhen);

  #-------------------------
  # Define local variables..
  #-------------------------
  
  my ( @sCmdOut, $iExitCode, @lsCmdout ) = ();
  my $sCommand = "$sBaseDir/pg_rewind";
  my $sCommandArgs = "-D $sDataDir --source-server=\"$sSourceConnStr\"";
  my $iTimeout = $iTillWhen - $^T;
  #----------------------
  # Set LOGNAME to user..
  #----------------------
  &logTrace ( 20395, "<%s> Executing Command [%s] with Args [%s]",
     $sMyName, $sCommand, $sCommandArgs);
  if ( PostgreSQL::ExecLocalCmd ( $sPostgreSQLUser,
                                  $sEnvFile,
                                  $sCommand,
                                  $sCommandArgs,
                                  \$iExitCode,
                                  $iTimeout,
                                  \@lsCmdout ) ) {
    &logTrace ( 20396,
        "<%s> Cmd [%s] with args [%s] executed successfully",
        $sMyName, $sCommand, $sCommandArgs );
    
    if ( 0 == $iExitCode ) { 
      &logTrace ( 20603,
                    "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                    $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
      if (-f "$sDataDir/recovery.done"){
        Sys::UnlinkFilesAsUser ( "$sDataDir/recovery.done" );
      }
      my $iExitCode = system ("$sCmdCP -p $sRecoveryFile $sDataDir"); 
      if ( $iExitCode == 0 ) {
        &logTrace ( 20463,
           "<%s> Cmd [%s] executed successfully",
           $sMyName, $sCmdCP);
      }
      else {
        &logError ( 20464,
           "<%s> Cmd [%s] failed to execute",
           $sMyName, $sCmdCP);
        return 0;
      } 
      if ($sRecoveryFile !~ /recovery.conf/) {
        my $sCommand = "$sCmdTouch";
        my $sCommandArgs = "$sDataDir/standby.signal";
        my $iTimeout = $iTillWhen - time;
        my @lsCmdout = ();
        
        if ( PostgreSQL::ExecLocalCmd ( $sPostgreSQLUser,
                                        "",
                                        $sCommand,
                                        $sCommandArgs,
                                        \$iExitCode,
                                        $iTimeout,
                                        \@lsCmdout ) ) {
          &logTrace ( 20617,
          "<%s> Cmd [%s] with args [%s] executed successfully",
          $sMyName, $sCommand, $sCommandArgs );
          if ( 0 == $iExitCode )
          {
                      &logTrace ( 20618,
                      "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                      $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
          }
          else
          {
            &logError ( 20619,
            "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
             $sMyName, $iExitCode, join ( chr 012, @lsCmdout ) );
             return 0 # Failure..
          }
        }
        else
        {
          &logError ( 20620,
          "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
          $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
          return 0; # Failure..
        }
      }
    }
    else { 
       &logError ( 20465,
          "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
           $sMyName, $iExitCode, join ( chr 012, @lsCmdout) );
         return 0;
    }
  }
  else {
    &logError ( 20604,
       "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
       $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
        return 0;
  }
  return 1; # Success
}


=pod
 -------------------------------------------------------------------------
 Function : int BaseBackup ( IN String DataDir,
                             IN String PostgreSQLUser,
                             IN String BackupCmd,
                             IN String RecoveryFile,
                             IN String EnvFile,
                             IN Integer TillWhen);

 Description : Uses PostgreSQL command pg_basebackup to connect as standby

 Parameters  :

 $sDataDir                  Location of the database storage area.
 $sPostgreSQLUser           Specifies the UserName required to run into pg_ctl.
 $sBackupCmd                Command used to base backup PostgreSQL data directory
 $sRecoveryFile             recovery.conf file provided to start standby instance
 $sEnvFile                  Absolute path to environment file to source
 $iTillWhen                 Time in EPOCH left for Operation

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut




sub BaseBackup {
  #----------------
  # Get Arguments..
  #----------------
  my($sDataDir, $sPostgreSQLUser, $sBackupCmd, $sRecoveryFile, $sEnvFile, $iTillWhen) = @_;
  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];
  
  &logTrace ( 20399, "Subroutine <%s> called with
  DataDir            [%s]
  PostgreSQLUser     [%s]
  BackupCmd          [%s]
  RecoveryFile       [%s]
  EnvFile            [%s]
  TillWhen           [%s]",
  $sMyName, $sDataDir, $sPostgreSQLUser, $sBackupCmd, $sRecoveryFile, $sEnvFile, $iTillWhen);

  #-------------------------
  # Define local variables..
  #-------------------------
  
  my (@sCmdOut, $iExitCode, @lsCmdout ) = ();
  
  my ($sCommand, $sCommandArgs) = $sBackupCmd
                                =~ /(\S+)\s+(.*)/; 
  
  #----------------------
  # Set LOGNAME to user..
  #----------------------
  my $sCmdToRun = "$sCommand $sCommandArgs";
  my $iTimeout = $iTillWhen - $^T;
  &logTrace ( 20402, "<%s> Executing Command [%s] with Args [%s]",
              $sMyName, $sCommand, $sCommandArgs);
  
   if ( PostgreSQL::ExecLocalCmd ( $sPostgreSQLUser,
                                   $sEnvFile,
                                   $sCommand,
                                   $sCommandArgs,
                                   \$iExitCode,
                                   $iTimeout,
                                   \@lsCmdout ) ) {
     &logTrace ( 20403,
        "<%s> Cmd [%s] with args [%s] executed successfully",
        $sMyName, $sCommand, $sCommandArgs );
    
     if ( 0 == $iExitCode ) { 
       &logTrace ( 20605,
                    "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                    $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
       if (PostgreSQL::IsRecoveryFileFound($sDataDir,
                                         $sRecoveryFile ) ){
         &logTrace ( 20454,"<%s>Recovery file found. Replacing it with file [%s].",
                    $sMyName, $sRecoveryFile );
         ( my $sFile = $sRecoveryFile ) =~ s#.*/(.*)$#$1#;
         Sys::UnlinkFilesAsUser ( "$sDataDir/$sFile" );
       }
       my $iExitCode = system ("$sCmdCP -p $sRecoveryFile $sDataDir");
       if ( $iExitCode == 0 ) {
           &logTrace ( 20606,
              "<%s> Cmd [%s] executed successfully",
              $sMyName, $sCmdCP);
        }
        else {
           &logError ( 20607,
              "<%s> Cmd [%s] failed to execute",
              $sMyName, $sCmdCP);
        return 0; 
        } 
       $iExitCode = system ("$sCmdTouch $sDataDir/standby.signal");
       if ( $iExitCode == 0 ) {
         &logTrace ( 20621,
           "<%s> Cmd [%s] executed successfully",
           $sMyName, $sCmdTouch);
       }
       else {
         &logError ( 20622,
           "<%s> Cmd [%s] failed to execute",
           $sMyName, $sCmdTouch);
         return 0;
       }
    }
    else { 
      &logError ( 20404,
         "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
          $sMyName, $iExitCode, join ( chr 012, @lsCmdout) );
         return 0;
    }
  }
  else {
     &logError ( 20608,
        "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
        $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
         return 0;
  }
  return 1; # Success
}


=pod
 -------------------------------------------------------------------------
 Function : int TakeOver ( IN String BaseDir,
                         IN String DataDir,
                         IN Integer Port,
                         IN String HostName,
                         IN String PostgreSQLUser,
                         IN String EnvFile,
                         IN String SourceConnStr,
                         IN String GlobalGrp,
                         IN String PGGroup,
                         IN String ResourceName,
                         IN Integer TillWhen);

 Description : Stops Old Master, promotes slave and checkpoints
 
 Parameters  :

 $sBaseDir                  The Installation Directory of PostgreSQL.
 $sDataDir                  Location of the database storage area.
 $iPort                     The port number on which PostgreSQL instance is configured.
 $sHostName                 Hostname/IPv4/IPv6 address of virtual host for PostgreSQL Instance
 $sPostgreSQLUser           Specifies the UserName required to run into pg_ctl.
 $sEnvFile                  Absolute path to environment file to source
 $sSourceConnStr            Connection string used by pg_rewind command
 $sGlobalGrp                Global Group used for Takeover Operation
 $sPGGroup                  PostgreSQL group to which the PostgreSQL resource belongs
 $bRResourceName            Current PostgreSQL ResourceName
 $iTillWhen                 Time in EPOCH left for Operation

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut




sub TakeOver {
  #----------------
  # Get Arguments..
  #----------------
  my($sBaseDir, $sDataDir, $iPort, $sHostName, $sPostgreSQLUser, $sEnvFile, $sSourceConnStr, $sGlobalGroup, $PGGroup, $sResourceName, $iTillWhen ) = @_;
  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];
  
  &logInfo ( 20536, "Subroutine <%s> called with
  BaseDir            [%s]
  DataDir            [%s]
  Port               [%s]
  HostName           [%s]
  PostgreSQLUser     [%s]
  EnvFile            [%s]
  SourceConnStr      [%s]
  GlobalGroup        [%s]
  PGGroup            [%s]
  ResourceName       [%s]
  TillWhen           [%s]",
  $sMyName, $sBaseDir, $sDataDir, $iPort, $sHostName, $sPostgreSQLUser, $sEnvFile, $sSourceConnStr, $sGlobalGroup, $PGGroup, $sResourceName, $iTillWhen  );

  
  #-------------------------
  # Define local variables..
  #-------------------------
  my ($sCommand, $sCommandArgs, $iExitCode, $sAttributeName, $Sys, $sOut,
      $sExpectedState, $iChildPid, $sFullCmd, $sCommandLine, $iTimeout ) = ();
  my (@lsOut, @lsCmdout) = ();
  my $bMasterRunning = 0;
  my $bExMasterFound = 0;

  ## Check if Master is running

  if (PostgreSQL::IsMasterRunning( $sBaseDir,
                                   $sPostgreSQLUser,
                                   $sEnvFile,
                                   $sSourceConnStr,
                                   $iTillWhen )){ 
    $bMasterRunning = 1;
    &logInfo ( 20537, "<%s> Master instance is running.",
          $sMyName );

    ## Stop Old Master

    my $sCurrentHost = "";
    if ( PostgreSQL::GetNodeId( \$sCurrentHost ) ){
      &logError ( 20538, "<%s>: Couldn't retrieve node name.", $sMyName );
      return 0;
    }
    else{
      &logInfo( 20539,"<%s>: Current host is [%s]", $sMyName, $sCurrentHost );
    }
   
    ## If Global group is unknown, find Global group from PGGroup
    if ($sGlobalGroup eq "" && $PGGroup ne "" ) { # global group unknown and PGGroup known
      $sCommand = $sCmdHagrp;
      $sCommandArgs = "-dep $PGGroup";;
      my $sGrp = "";
      my $iExitCode = 0;
      $iTimeout = $iTillWhen - $^T;
      &logInfo ( 20540, "<%s> Executing Command [%s] with arguments [%s]",
              $sMyName, $sCommand, $sCommandArgs );
      if ( PostgreSQL::ExecLocalCmd ( "",
                                      "",
                                      $sCommand,
                                      $sCommandArgs,
                                      \$iExitCode,
                                      $iTimeout,
                                      \@lsCmdout ) ) {
        &logTrace ( 20541,
        "<%s> Cmd [%s] with args [%s] executed successfully",
        $sMyName, $sCommand, $sCommandArgs );
    
        if ( 0 == $iExitCode )
        { 
          &logTrace ( 20542,
                    "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                    $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
          my (@lsOut, $sOut) = ();
          foreach $sOut (@lsCmdout) {
            @lsOut = split( '\s+', $sOut);
            if( $lsOut[1] eq $PGGroup ) {
              $sGlobalGroup = $lsOut[0];
              last;
            }
          }
        }
        else
        { 
          &logError ( 20543,
          "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
           $sMyName, $iExitCode, join ( chr 012, @lsCmdout) );
           return 0 # Failure..
        }
      }
      else
      {
        &logError ( 20544,
        "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
        $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
        return 0; # Failure..
      }
    }  # end if to find global group name

    # Now we know the TakeOver Group name
    # GlobalGroup

    #-----------------------------------------------------------------------
    # Find the State of the Global Group. 
    # 
    # Sample output:
    #
    # #hagrp -state TakeOver_SG
    # #Group       Attribute             System               Value
    # TakeOver_SG  State                 localclus:vmrac020   |OFFLINE|
    # TakeOver_SG  State                 pg1:vmrac018         |ONLINE|
    # TakeOver_SG  State                 pg1:vmrac019         |OFFLINE|
    #                             OR
    # #hagrp -state TakeOver_SG
    # #Group       Attribute             System     Value
    # TakeOver_SG  State                 vcsesx001  |OFFLINE|
    # TakeOver_SG  State                 vcsesx002  |ONLINE|
    #                             OR
    #
    # #hagrp -state TakeOver_SG
    # #Group       Attribute             System               Value
    # TakeOver_SG  State                 localclus:vmrac020   |OFFLINE|
    # TakeOver_SG  State                 pg1:vmrac018         |ONLINE|
    #                             OR
    # #hagrp -state TakeOver_SG
    # #Group       Attribute             System                    Value
    # TakeOver_SG  State                 localclus:dl380g9-49-v012 |OFFLINE|
    # TakeOver_SG  State                 localclus:dl380g9-49-v013 |OFFLINE|
    # TakeOver_SG  State                 pgclus1:dl380g9-49-v010   |OFFLINE|
    # TakeOver_SG  State                 pgclus1:dl380g9-49-v011   |OFFLINE|
    #                             OR
    # TakeOver_SG  State                 localclus:dl380g9-49-v012 |OFFLINE|
    # TakeOver_SG  State                 localclus:dl380g9-49-v013 |OFFLINE|
    # TakeOver_SG  State                 pgclus1:dl380g9-49-v010   |OFFLINE|
    #
    #------------------------------------------------------------------------
    
    $sCommand = $sCmdHagrp;
    $sCommandArgs = "-state $sGlobalGroup";
    $iTimeout = $iTillWhen - $^T;
    my @lsSysList = ();  # List of the systems to check for online PG SG
    &logInfo ( 20545, "<%s> Executing Command [%s] with arguments [%s]",
                $sMyName, $sCommand, $sCommandArgs );
    if ( PostgreSQL::ExecLocalCmd ( "",
                                    "",
                                    $sCommand,
                                    $sCommandArgs,
                                    \$iExitCode,
                                    $iTimeout,
                                    \@lsCmdout ) ) {
      &logTrace ( 20546,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
  
      if ( 0 == $iExitCode )
      {
        &logTrace ( 20547,
                  "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                  $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
        shift @lsCmdout;
        my $iNumSys = scalar @lsCmdout;
        my ($sCluster1, $sSystem1) = ();
    
        foreach $sOut (@lsCmdout) {
          @lsOut = split( '\s+', $sOut);
          if ( $lsOut[2] =~ /:/) {
           ($sCluster1, $sSystem1) = split(':',$lsOut[2]);
          }
          else {
            $sCluster1 = "";
            $sSystem1 = $lsOut[2];
          }
          if ($sSystem1 eq $sCurrentHost) {
            next;
          }
          if( $sCluster1 ne "localclus" ) {
              if (PostgreSQL::CheckSSH($sSystem1,$iTillWhen)) {
              &logInfo ( 20548, "<%s> Agent will check PostgreSQL SG state on the System [%s] in cluster [%s]",
                    $sMyName, $sSystem1, $sCluster1 );
              push ( @lsSysList, $sSystem1 );
            }
          }
        }
      }
      else
      {
        &logError ( 20549,
        "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
         $sMyName, $iExitCode, join ( chr 012, @lsCmdout) );
         return 0 # Failure..
      }
    }
    else
    {
      &logError ( 20550,
      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
      return 0; # Failure..
    }

    foreach $Sys (@lsSysList) {
      my $SysState = "";    
      #### check if vcs is running on other node.
      my $iRetCode = PostgreSQL::GetSysState($Sys, \$SysState);
      if ($SysState ne "RUNNING"){
      ### vcs not active on other(possibly master) node.
          &logInfo ( 20551, "<%s> VCS is not running on System [%s].",
                $sMyName, $Sys );
        next;
      }
      else {
      ## VCS is running on this node
      &logInfo ( 20552, "<%s> VCS is running on System [%s]. Agent will now check the status of PostgreSQL on this node.",
                 $sMyName, $Sys );
      }
      
      
      #--------------------------------------------------------------
      # Find the state of PostgreSQL group
      # 
      # Sample output:
      #
      # #ssh vmrac018 '/opt/VRTSvcs/bin/hagrp -value PG_SG State'
      # |ONLINE|
      #
      # #ssh vcsesx001 '/opt/VRTSvcs/bin/hagrp -value PG_SG State'
      # |OFFLINE|FAULTED|
      #
      #---------------------------------------------------------------
      $sCommand = $sCmdSSH;
      $sCommandArgs = " $Sys \'$HAGRP -value $PGGroup State\'";
      $iTimeout = $iTillWhen - $^T;
      &logInfo ( 20553, "<%s> Executing Command [%s] with arguments [%s]",
                  $sMyName, $sCommand, $sCommandArgs );
      if ( PostgreSQL::ExecLocalCmd ( "",
                                      "",
                                      $sCommand,
                                      $sCommandArgs,
                                      \$iExitCode,
                                      $iTimeout,
                                      \@lsCmdout ) ) {
        &logTrace ( 20554,
        "<%s> Cmd [%s] with args [%s] executed successfully",
        $sMyName, $sCommand, $sCommandArgs );
    
        if ( 0 == $iExitCode ) {
          &logTrace ( 20555,
                    "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                    $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
          chomp @lsCmdout;
          &logInfo ( 20556, "<%s> Agent found Group [%s] state [%s] on System [%s] ",
                  $sMyName, $PGGroup, $lsCmdout[0], $Sys );
      
          if ($lsCmdout[0] eq "|ONLINE|"){
            $bExMasterFound = 1;
            $sCommand = $sCmdSSH;
            $sCommandArgs = " $Sys \'$HAGRP -value $sGlobalGroup State\'";
            $iTimeout = $iTillWhen - $^T;
            &logInfo ( 20584, "<%s> Executing Command [%s] with arguments [%s]",
                        $sMyName, $sCommand, $sCommandArgs );
            if ( PostgreSQL::ExecLocalCmd ( "",
                                            "",
                                            $sCommand,
                                            $sCommandArgs,
                                            \$iExitCode,
                                            $iTimeout,
                                            \@lsCmdout ) ) {
              &logTrace ( 20585,
              "<%s> Cmd [%s] with args [%s] executed successfully",
              $sMyName, $sCommand, $sCommandArgs );
          
              if ( 0 == $iExitCode ) {
                &logTrace ( 20586,
                          "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                          $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
                chomp @lsCmdout;
                &logInfo ( 20587, "<%s> Agent found Group [%s] state [%s] on System [%s] ",
                        $sMyName, $sGlobalGroup, $lsCmdout[0], $Sys );
            
                if ($lsCmdout[0] eq "|ONLINE|"){
                  $sExpectedState =  "|OFFLINE|";
                  #---------------------------------------------------------------------
                  # Offline PostgreSQL group of Master, if online
                  # 
                  # Sample command:
                  # #ssh vmrac018 '/opt/VRTSvcs/bin/hagrp -offline PG_SG -sys vmrac018'
                  #----------------------------------------------------------------------
                  &logInfo ( 20588, "<%s> Agent found Group [%s] state as [%s], going to OFFLINE group on System [%s] ExpectedState is [%s] ",
                        $sMyName, $sGlobalGroup, $lsCmdout[0], $Sys, $sExpectedState );
            
                  if ( PostgreSQL::RemoteGroupOperation( $sGlobalGroup,
                                                    "offline",
                                                    $Sys,
                                                    $sExpectedState,
                                                    $iTillWhen ) ) {
                    &logInfo ( 20589,
                                "<%s> Remote group Operation successful on group[%s] system [%s].",
                                $sMyName, $sGlobalGroup, $Sys ); 
                  }
                  else {
                    &logError ( 20590,
                                "<%s>Remote group Operation failed on group[%s] system [%s].",
                                $sMyName, $sGlobalGroup, $Sys);
                  }
                }
                else {
                  &logInfo ( 20591, "<%s> Agent found Group [%s] state as [%s]",
                        $sMyName, $sGlobalGroup, $lsCmdout[0] );
                }
              }
              else
              {
                &logError ( 20592,
                "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
                 $sMyName, $iExitCode, join ( chr 012, @lsCmdout) );
                 return 0 # Failure..
              }
            }
            else
            {
              &logError ( 20593,
              "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
              $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
              return 0; # Failure..
            }
            $sExpectedState =  "|OFFLINE|";
            #---------------------------------------------------------------------
            # Offline PostgreSQL group of Master, if online
            # 
            # Sample command:
            # #ssh vmrac018 '/opt/VRTSvcs/bin/hagrp -offline PG_SG -sys vmrac018'
            #----------------------------------------------------------------------
            &logInfo ( 20557, "<%s> Agent found Group [%s] state as [%s], going to OFFLINE group on System [%s] ExpectedState is [%s] ",
                  $sMyName, $PGGroup, $lsCmdout[0], $Sys, $sExpectedState );
      
            if ( PostgreSQL::RemoteGroupOperation( $PGGroup,
                                              "offline",
                                              $Sys,
                                              $sExpectedState,
                                              $iTillWhen ) ) {
              &logInfo ( 20558,
                          "<%s> Remote group Operation successful on group[%s] system [%s].",
                          $sMyName, $PGGroup, $Sys ); 
            }
            else {
              &logError ( 20559,
                          "<%s>Remote group Operation failed on group[%s] system [%s].",
                          $sMyName, $PGGroup, $Sys);
            }
            last;
          }
          else {
            &logInfo ( 20560, "<%s> Agent found Group [%s] state as [%s]",
                  $sMyName, $PGGroup, $lsCmdout[0] );
          }
        }
        else
        {
          &logError ( 20561,
          "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
           $sMyName, $iExitCode, join ( chr 012, @lsCmdout) );
           return 0 # Failure..
        }
      }
      else
      {
        &logError ( 20562,
        "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
        $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
        return 0; # Failure..
      }
    } # for end
  }
  else {
    # Consider offline.
    &logInfo ( 20563, "<%s> Master instance is not running.",
           $sMyName );
  }
  #### Log if possible master-master scenario.

  if ($bMasterRunning == 1 && $bExMasterFound == 0) {
  #### Quit 
    &logError ( 20564, "<%s> Master Instance is running, but VCS may not be running on Master node. Manually stop the master instance to avoid Master-Master scenario.",
      $sMyName );
    return 0;
  }
  #### Promote
  
  sleep (1);
  $sCommand = "$sBaseDir/pg_ctl";
  $sCommandArgs = "promote -w -D $sDataDir -o \"-p $iPort \" ";
  $iTimeout = $iTillWhen - $^T; 
  &logInfo ( 20601, "<%s> Executing Command [%s] with arguments [%s]",
                  $sMyName, $sCommand, $sCommandArgs );
  if ( PostgreSQL::ExecLocalCmd ( $sPostgreSQLUser,
                                  $sEnvFile,
                                  $sCommand,
                                  $sCommandArgs,
                                  \$iExitCode,
                                  $iTimeout,
                                  \@lsCmdout ) ) {
    &logTrace ( 20565,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  
    if ( 0 == $iExitCode )
    { 
      &logTrace ( 20566,
                "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
    }
    else
    { 
      &logError ( 20567,
      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
       $sMyName, $iExitCode, join ( chr 012, @lsCmdout) );
       return 0 # Failure..
    }
  }
  else
  {
    &logError ( 20568,
    "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
    return 0; # Failure..
  }
  
  ### checkpoint
  $sCommand = "$sBaseDir/psql";
  $sCommandArgs = "-A -t -h $sHostName -p $iPort -c \"checkpoint\"";
  $iTimeout = $iTillWhen - $^T;
  &logInfo ( 20602, "<%s> Executing Command [%s] with arguments [%s]",
                  $sMyName, $sCommand, $sCommandArgs );
  if ( PostgreSQL::ExecLocalCmd ( $sPostgreSQLUser,
                                  $sEnvFile,
                                  $sCommand,
                                  $sCommandArgs,
                                  \$iExitCode,
                                  $iTimeout,
                                  \@lsCmdout ) ) {
    &logTrace ( 20569,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  
    if ( 0 == $iExitCode )
    { 
      &logTrace ( 20570,
                "<%s> Command [%s] with Args [%s] executed successfully with output [%s]",
                $sMyName, $sCommand, $sCommandArgs, join ( chr 012, @lsCmdout ) );
    }
    else
    { 
      &logError ( 20571,
      "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
       $sMyName, $iExitCode, join ( chr 012, @lsCmdout) );
       return 0 # Failure..
    }
  }
  else
  {
    &logError ( 20572,
    "<%s> Command [%s] with Args [%s] failed with Exit Code [%s] and output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
    return 0; # Failure..
  }
return 1;
}

